@@ -28,8 +28,13 @@ extern "C"
 #include <libswscale/swscale.h>
 }
 
-#include <AL/al.h>
-#include <AL/alc.h>
+#ifdef __APPLE__
+#  include <OpenAL/al.h>
+#  include <OpenAL/alc.h>
+#else
+#  include <AL/al.h>
+#  include <AL/alc.h>
+#endif
 
 #include <GL/glew.h>
 
