@@ -108,9 +108,13 @@ public:
         get_media_input_nonconst().seek(pos);
     }
 
-    void read_frame()   // Only called on slave nodes to force reading the next frame
+    void start_frame_read()     // Only called on slave nodes
     {
         get_media_input_nonconst().start_video_frame_read();
+    }
+
+    void finish_frame_read()    // Only called on slave nodes
+    {
         _video_frame = get_media_input_nonconst().finish_video_frame_read();
         if (!_video_frame.is_valid())
         {
@@ -180,6 +184,51 @@ public:
     }
 };
 
+/*
+ * Define errors produced by player.
+ */
+
+enum Error
+{
+    ERROR_MAP_INITDATA_FAILED = eq::ERROR_CUSTOM,
+    ERROR_MAP_FRAMEDATA_FAILED,
+    ERROR_PLAYER_INIT_FAILED,
+    ERROR_OPENGL_2_1_NEEDED
+};
+
+struct ErrorData
+{
+    const uint32_t code;
+    const std::string text;
+};
+
+static ErrorData errors[] =
+{
+    { ERROR_MAP_FRAMEDATA_FAILED, "Init data mapping failed" },
+    { ERROR_MAP_INITDATA_FAILED, "Frame data mapping failed" },
+    { ERROR_PLAYER_INIT_FAILED, "Video player initialization failed" },
+    { ERROR_OPENGL_2_1_NEEDED, "Need at least OpenGL 2.1" },
+    { 0, "" } // last!
+};
+
+static void initErrors()
+{
+    co::base::ErrorRegistry &registry = co::base::Global::getErrorRegistry();
+    for (size_t i = 0; errors[i].code != 0; i++)
+    {
+        registry.setString(errors[i].code, errors[i].text);
+    }
+}
+
+static void exitErrors()
+{
+    co::base::ErrorRegistry &registry = co::base::Global::getErrorRegistry();
+    for (size_t i = 0; errors[i].code != 0; i++)
+    {
+        registry.eraseString(errors[i].code);
+    }
+}
+
 /*
  * eq_init_data
  */
@@ -622,16 +671,17 @@ public:
     }
 
 protected:
-    virtual bool configInit(const eq::uint128_t& initID)
+    virtual bool configInit(const eq::uint128_t &init_id)
     {
-        if (!eq::Node::configInit(initID))
+        if (!eq::Node::configInit(init_id))
         {
             return false;
         }
         // Map our InitData instance to the master instance
         eq_config *config = static_cast<eq_config *>(getConfig());
-        if (!config->mapObject(&init_data, initID))
+        if (!config->mapObject(&init_data, init_id))
         {
+            setError(ERROR_MAP_INITDATA_FAILED);
             return false;
         }
         // Is this the application node?
@@ -642,6 +692,7 @@ protected:
         // Map our FrameData instance to the master instance
         if (!config->mapObject(&frame_data, init_data.frame_data_id))
         {
+            setError(ERROR_MAP_FRAMEDATA_FAILED);
             return false;
         }
 
@@ -652,6 +703,7 @@ protected:
         {
             if (!_player.init(init_data.init_data, frame_template))
             {
+                setError(ERROR_PLAYER_INIT_FAILED);
                 return false;
             }
         }
@@ -677,10 +729,10 @@ protected:
         return eq::Node::configExit();
     }
 
-    virtual void frameStart(const eq::uint128_t& frameID, const uint32_t frameNumber)
+    virtual void frameStart(const eq::uint128_t &frame_id, const uint32_t frame_number)
     {
         // Update our frame data
-        frame_data.sync(frameID);
+        frame_data.sync(frame_id);
         // Do as we're told
         if (_is_app_node)
         {
@@ -691,17 +743,37 @@ protected:
             if (frame_data.seek_to >= 0)
             {
                 _player.seek(frame_data.seek_to);
+                _player.start_frame_read();
             }
             if (frame_data.prep_frame)
             {
-                _player.read_frame();
+                _player.finish_frame_read();
             }
             if (frame_data.drop_frame)
             {
-                _player.read_frame();
+                _player.finish_frame_read();
+                _player.start_frame_read();
             }
         }
-        startFrame(frameNumber);
+        startFrame(frame_number);
+    }
+
+    virtual void frameFinish(const eq::uint128_t &, const uint32_t frame_number)
+    {
+        if (_is_app_node)
+        {
+            // Nothing to do since the config's master player already did it
+        }
+        else
+        {
+            if (frame_data.prep_frame)
+            {
+                // The frame was uploaded to texture memory.
+                // Start reading the next one asynchronously.
+                _player.start_frame_read();
+            }
+        }
+        releaseFrame(frame_number);
     }
 
 public:
@@ -751,10 +823,10 @@ public:
     }
 
 protected:
-    virtual bool configInitGL(const eq::uint128_t& initID)
+    virtual bool configInitGL(const eq::uint128_t &init_id)
     {
         msg::dbg(HERE);
-        if (!eq::Window::configInitGL(initID))
+        if (!eq::Window::configInitGL(init_id))
         {
             return false;
         }
@@ -762,6 +834,7 @@ protected:
                     "GL_VERSION_2_1 GL_EXT_framebuffer_object"))
         {
             msg::err("This OpenGL implementation does not support OpenGL 2.1 and framebuffer objects.");
+            setError(ERROR_OPENGL_2_1_NEEDED);
             return false;
         }
 
@@ -780,7 +853,7 @@ protected:
         return eq::Window::configExitGL();
     }
 
-    virtual void frameStart(const eq::uint128_t&, const uint32_t frameNumber)
+    virtual void frameStart(const eq::uint128_t &, const uint32_t frame_number)
     {
         // Get frame data via from the node
         eq_node *node = static_cast<eq_node *>(getNode());
@@ -794,12 +867,12 @@ protected:
         {
             _video_output.activate_next_frame();
         }
-        startFrame(frameNumber);
+        startFrame(frame_number);
     }
 
-    virtual void frameFinish(const eq::uint128_t&, const uint32_t frameNumber)
+    virtual void frameFinish(const eq::uint128_t &, const uint32_t frame_number)
     {
-        releaseFrame(frameNumber);
+        releaseFrame(frame_number);
     }
 };
 
@@ -817,10 +890,10 @@ public:
     }
 
 protected:
-    virtual void frameDraw(const eq::uint128_t& frameID)
+    virtual void frameDraw(const eq::uint128_t &frame_id)
     {
         // Let Equalizer initialize some stuff
-        eq::Channel::frameDraw(frameID);
+        eq::Channel::frameDraw(frame_id);
 
         // Get the canvas video area and the canvas channel area
         eq_node *node = static_cast<eq_node *>(getNode());
@@ -906,6 +979,7 @@ player_equalizer::player_equalizer(int *argc, char *argv[], bool flat_screen) :
     player(player::slave), _flat_screen(flat_screen)
 {
     /* Initialize Equalizer */
+    initErrors();
     _node_factory = static_cast<void *>(new eq_node_factory);
     if (!eq::init(*argc, argv, static_cast<eq::NodeFactory *>(_node_factory)))
     {
@@ -946,6 +1020,7 @@ void player_equalizer::run()
     config->exit();
     eq::releaseConfig(config);
     eq::exit();
+    exitErrors();
 }
 
 void player_equalizer::close()
