@@ -1251,7 +1251,7 @@ main_window::main_window(QSettings *settings, const player_init_data &init_data)
 {
     // Application properties
     setWindowTitle(PACKAGE_NAME);
-    setWindowIcon(QIcon(":icons/appicon.png"));
+    setWindowIcon(QIcon(":logo/bino_logo_small_64x64.png"));
 
     // Load preferences
     _settings->beginGroup("Session");
@@ -1601,18 +1601,19 @@ void main_window::playloop_step()
     }
     else
     {
-        bool r = false;
         try
         {
-            r = _player->playloop_step();
+            if (!_player->playloop_step())
+            {
+                _timer->stop();
+            }
         }
         catch (std::exception &e)
-        {
-            QMessageBox::critical(this, "Error", e.what());
-        }
-        if (!r)
         {
             _timer->stop();
+            send_cmd(command::toggle_play);
+            _player->playloop_step();   // react on command
+            QMessageBox::critical(this, "Error", e.what());
         }
     }
 }
@@ -1806,7 +1807,7 @@ void main_window::help_about()
     dialog->setWindowTitle("About " PACKAGE_NAME);
 
     QLabel *logo = new QLabel();
-    logo->setPixmap(QPixmap(":icons/appicon.png").scaledToHeight(32, Qt::SmoothTransformation));
+    logo->setPixmap(QPixmap(":logo/bino_logo.png").scaledToHeight(128, Qt::SmoothTransformation));
     QLabel *spacer = new QLabel();
     spacer->setMinimumWidth(16);
     QLabel *title = new QLabel("The " PACKAGE_NAME " 3D video player");
@@ -1825,6 +1826,7 @@ void main_window::help_about()
 
     QTextBrowser *libs_text = new QTextBrowser(this);
     libs_text->setOpenExternalLinks(true);
+    libs_text->setWordWrapMode(QTextOption::NoWrap);
     QString libs_blurb = tr("<p>Platform:<ul><li>%1</li></ul></p>").arg(PLATFORM);
     libs_blurb += QString("<p>Libraries used:");
     std::vector<std::string> libs = lib_versions(true);
@@ -1835,9 +1837,25 @@ void main_window::help_about()
     libs_blurb += QString("</p>");
     libs_text->setText(libs_blurb);
 
+    QTextBrowser *team_text = new QTextBrowser(this);
+    team_text->setOpenExternalLinks(false);
+    team_text->setWordWrapMode(QTextOption::NoWrap);
+    QFile team_file(":AUTHORS");
+    team_file.open(QIODevice::ReadOnly | QIODevice::Text);
+    team_text->setText(team_file.readAll());
+
+    QTextBrowser *license_text = new QTextBrowser(this);
+    license_text->setOpenExternalLinks(false);
+    license_text->setWordWrapMode(QTextOption::NoWrap);
+    QFile license_file(":LICENSE");
+    license_file.open(QIODevice::ReadOnly | QIODevice::Text);
+    license_text->setText(license_file.readAll());
+
     QTabWidget *tab_widget = new QTabWidget();
     tab_widget->addTab(about_text, "About");
     tab_widget->addTab(libs_text, "Libraries");
+    tab_widget->addTab(team_text, "Team");
+    tab_widget->addTab(license_text, "License");
 
     QPushButton *ok_btn = new QPushButton("OK");
     connect(ok_btn, SIGNAL(pressed()), dialog, SLOT(accept()));
