@@ -20,23 +20,54 @@
 
 #include "config.h"
 
+#include <cstdlib>
+
 #include <QCoreApplication>
 #include <QApplication>
+#include <QtGlobal>
+#include <QTextCodec>
 
 #include "qt_app.h"
 
 #include "msg.h"
 
 
-static int qt_argc = 1;
-static const char *qt_argv[2] = { PACKAGE_NAME, NULL };
+static int qt_argc = 0;
+static char **qt_argv = NULL;
 static QApplication *qt_app = NULL;
 
+void set_qt_argv(int argc, char *argv[])
+{
+    qt_argc = argc;
+    qt_argv = argv;
+}
+
+static void qt_msg_handler(QtMsgType type, const char *msg)
+{
+    switch (type)
+    {
+    case QtDebugMsg:
+        msg::dbg("%s", msg);
+        break;
+    case QtWarningMsg:
+        msg::wrn("%s", msg);
+        break;
+    case QtCriticalMsg:
+        msg::err("%s", msg);
+        break;
+    case QtFatalMsg:
+        msg::err("%s", msg);
+        std::abort();
+    }
+}
+
 bool init_qt()
 {
     if (!qt_app)
     {
+        qInstallMsgHandler(qt_msg_handler);
         qt_app = new QApplication(qt_argc, const_cast<char **>(qt_argv));
+        QTextCodec::setCodecForCStrings(QTextCodec::codecForName("UTF-8"));
         QCoreApplication::setOrganizationName("Bino");
         QCoreApplication::setOrganizationDomain("bino.nongnu.org");
         QCoreApplication::setApplicationName(PACKAGE_NAME);
