@@ -55,7 +55,6 @@ video_output_qt_widget::video_output_qt_widget(
     QGLWidget(format, parent), _vo(vo)
 {
     setFocusPolicy(Qt::StrongFocus);
-    setWindowIcon(QIcon(":icons/appicon.png"));
 }
 
 video_output_qt_widget::~video_output_qt_widget()
@@ -199,6 +198,7 @@ void video_output_qt_widget::mouseReleaseEvent(QMouseEvent *event)
 
 video_container_widget::video_container_widget(QWidget *parent) : QWidget(parent), _w(64), _h(64)
 {
+    setWindowIcon(QIcon(":logo/bino_logo_small_64x64.png"));
     // Set minimum size > 0 so that the container is always visible
     setMinimumSize(_w, _h);
     // Always paint the complete widget black
@@ -392,7 +392,7 @@ void video_output_qt::mouse_set_pos(float dest)
         // Disabled in fullscreen and GUI mode
         return;
     }
-    if (_playing)
+    if (_playing || !_container_is_external)
     {
         send_cmd(command::set_pos, dest);
     }
